import dataclasses
import datetime
import re
import typing

from pydantic_core import CoreSchema

from pydantic import Field, GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue

from .schema import BaseModel, Enum


#: Regex used for extracting versions from module names
VERSION_REGEX = re.compile(r"^v\d+((alpha|beta)\d+)?$")


class Scope(str, Enum):
    """
    Enum representing the possible scopes for a custom resource.
    """
    CLUSTER    = "Cluster"
    NAMESPACED = "Namespaced"


@dataclasses.dataclass
class CustomResourceMeta:
    """
    Dataclass to hold metadata for the custom resource.
    """
    # The API subgroup for the custom resource
    api_subgroup: typing.Optional[str]
    # The version of the custom resource
    version: str
    # The scope of the custom resource
    scope: Scope
    # The kind of the resource
    kind: str
    # The singular name of the resource
    singular_name: str
    # The plural name of the resource
    plural_name: str
    # A list of short names for the resource
    short_names: typing.List[str]
    # The subresources for the CRD
    subresources: typing.Dict[str, typing.Any]
    # A list of printer column definitions for the version
    printer_columns: typing.List[typing.Dict[str, typing.Any]]
    # Indicates if this is the storage version
    storage_version: bool


class CustomResourceMetaclass(type(BaseModel)):
    """
    Metaclass for the custom resource class.
    """
    def __new__(
        cls,
        name,
        bases,
        attrs,
        /,
        # Indicates that the resource is abstract and is not actually a CRD
        abstract: bool = False,
        # The API subgroup for the custom resource
        # If not given, the resource is placed in the root api group
        api_subgroup: typing.Optional[str] = None,
        # The version of the custom resource
        # If not given, it is inferred from the module if possible
        version: typing.Optional[str] = None,
        # The scope of the resource, either 'Namespaced' or 'Cluster'
        # If not given, Namespaced is used by default
        scope: Scope = Scope.NAMESPACED,
        # The kind of the resource
        # If not given, the model class name is used
        kind: typing.Optional[str] = None,
        # The singular name of the resource
        # If not given, the lower-cased kind is used
        singular_name: typing.Optional[str] = None,
        # The plural name of the resource
        # If not given, the singular name with 's' appended is used
        plural_name: typing.Optional[str] = None,
        # A list of short names for the resource
        # If not given, no short names are used
        short_names: typing.Optional[typing.List[str]] = None,
        # The subresources for the CRD
        # If not given, no subresources are defined
        subresources: typing.Optional[typing.Dict[str, typing.Any]] = None,
        # A list of printer column definitions for the version
        # An "age" column is always appended to the given list
        printer_columns: typing.Optional[typing.List[typing.Dict[str, typing.Any]]] = None,
        # Indicates if this is the storage version
        storage_version: bool = True,
        # Other kwargs, passed to Pydantic
        **kwargs
    ):
        if not abstract:
            # If no version is given, find the first module component that looks like a version
            if not version:
                version = next(
                    part
                    for part in reversed(attrs["__module__"].split("."))
                    if VERSION_REGEX.match(part)
                )
            # Derive any names that were't given from the ones that were
            kind = kind or name
            singular_name = singular_name or kind.lower()
            plural_name = plural_name or f"{singular_name}s"
            # Add the metadata to the attrs
            # We also add an annotation so that Pydantic leaves it alone
            attrs.setdefault("__annotations__", {})["_meta"] = typing.ClassVar[CustomResourceMeta]
            attrs["_meta"] = CustomResourceMeta(
                api_subgroup,
                version,
                scope,
                kind,
                singular_name,
                plural_name,
                short_names or [],
                subresources or {},
                (printer_columns or []) + [
                    {
                        "name": "Age",
                        "type": "date",
                        "jsonPath": ".metadata.creationTimestamp",
                    },
                ],
                storage_version
            )
        return super().__new__(cls, name, bases, attrs, **kwargs)


class OwnerReference(BaseModel):
    """
    Model for an owner reference.
    """
    api_version: str = Field(
        ...,
        description = "The API version of the owner."
    )
    kind: str = Field(
        ...,
        description = "The kind of the owner."
    )
    name: str = Field(
        ...,
        description = "The name of the owner."
    )
    uid: str = Field(
        ...,
        description = "The UID of the owner."
    )
    block_owner_deletion: bool = Field(
        False,
        description = "If true, the owner cannot be removed until this resource is removed."
    )
    controller: bool = Field(
        False,
        description = "Indicates if this resource points to the managing controller."
    )


class Metadata(BaseModel):
    """
    Model for the metadata associated with a resource instance.
    """
    name: str = Field(
        ...,
        description = "The name of the resource."
    )
    namespace: typing.Optional[str] = Field(
        None,
        description = "The namespace for the resource."
    )
    labels: typing.Dict[str, str] = Field(
        default_factory = dict,
        description = "The labels for the resource. Can be used to match selectors."
    )
    annotations: typing.Dict[str, str] = Field(
        default_factory = dict,
        description = "Annotations for the resource. Can be used to store arbitrary metadata."
    )
    owner_references: typing.List[OwnerReference] = Field(
        default_factory = list,
        description = "List of resources that this resource depends on."
    )
    finalizers: typing.List[str] = Field(
        default_factory = list,
        description = "List of identifiers blocking removal of the resource."
    )
    uid: typing.Optional[str] = Field(
        None,
        description = "The UID for the resource."
    )
    creation_timestamp: typing.Optional[datetime.datetime] = Field(
        None,
        description = "The timestamp at which the resource was created."
    )
    deletion_timestamp: typing.Optional[datetime.datetime] = Field(
        None,
        description = "The timestamp at which the resource was deleted."
    )
    resource_version: typing.Optional[str] = Field(
        None,
        description = (
            "The internal version of the resource. "
            "Used to implement optimistic concurrency and resumable watches."
        )
    )

    def add_owner_reference(self, owner, /, block_owner_deletion = False, controller = False):
        """
        Adds the given owner as a reference if it doesn't already exist.

        Returns True if a new reference was added, False otherwise.
        """
        if any(ref.uid == owner.metadata.uid for ref in self.owner_references):
            # Reference already exists - nothing to do
            return False
        else:
            self.owner_references.append(
                OwnerReference(
                    api_version = owner["apiVersion"],
                    kind = owner["kind"],
                    name = owner["metadata"]["name"],
                    uid = owner["metadata"]["uid"],
                    block_owner_deletion = block_owner_deletion,
                    controller = controller
                )
            )
            return True


class CustomResource(
    BaseModel,
    metaclass = CustomResourceMetaclass,
    abstract = True
):
    """
    Base class for defining custom resources.
    """
    api_version: str = Field(
        ...,
        description = "The API version of the resource."
    )
    kind: str = Field(
        ...,
        description = "The kind of the resource."
    )
    metadata: Metadata = Field(
        ...,
        description = "The metadata for the resource."
    )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = super().__get_pydantic_json_schema__(core_schema, handler)
        # Remove the API version, kind and metadata from the schema as they are
        # not required for a valid CRD
        for key in ["apiVersion", "kind", "metadata"]:
            json_schema["properties"].pop(key)
            json_schema["required"].remove(key)
        return json_schema



import annotated_types as a
import typing as t

import pydantic as p

from kube_custom_resource import custom_resource as crd, schema as s


ModelsList = t.Annotated[t.List[s.constr(pattern = r"^[a-z0-9]+$")], a.Len(min_length = 1)]


class ManufacturerSpec(s.BaseModel):
    """Spec for a car manufacturer."""
    founded: s.IntOrString = p.Field(
        ...,
        description = "The year the company was founded."
    )
    website: s.AnyHttpUrl = p.Field(
        ...,
        description = "The company website."
    )
    models: ModelsList = p.Field(
        ...,
        description = "The models supported by this manufacturer."
    )


class Manufacturer(
    crd.CustomResource,
    scope = crd.Scope.CLUSTER,
    subresources = {"status": {}},
    printer_columns = [
        {
            "name": "Founded",
            "type": "string",
            "jsonPath": ".spec.founded",
        },
    ],
):
    """Represents a car manufacturer."""
    spec: ManufacturerSpec



import kube_custom_resource as kcr

import models


API_GROUP = "test.kcr.azimuth-cloud.io"


registry = kcr.CustomResourceRegistry(API_GROUP)
registry.discover_models(models)


def test_model_discovery_found_two_models():
    assert len(registry) == 2


MANUFACTURER_CRD_EXPECTED = {
    "apiVersion": "apiextensions.k8s.io/v1",
    "kind": "CustomResourceDefinition",
    "metadata": {
        "name": "manufacturers.test.kcr.azimuth-cloud.io"
    },
    "spec": {
        "group": "test.kcr.azimuth-cloud.io",
        "scope": "Cluster",
        "names": {
            "kind": "Manufacturer",
            "singular": "manufacturer",
            "plural": "manufacturers",
            "shortNames": [],
            "categories": []
        },
        "versions": [
            {
                "name": "v1alpha1",
                "served": True,
                "storage": True,
                "schema": {
                    "openAPIV3Schema": {
                        "description": "Represents a car manufacturer.",
                        "properties": {
                            "spec": {
                                "description": "Spec for a car manufacturer.",
                                "properties": {
                                    "founded": {
                                        "description": "The year the company was founded.",
                                        "x-kubernetes-int-or-string": True
                                    },
                                    "website": {
                                        "description": "The company website.",
                                        "format": "uri",
                                        "minLength": 1,
                                        "type": "string"
                                    },
                                    "models": {
                                        "description": "The models supported by this manufacturer.",
                                        "items": {
                                            "pattern": "^[a-z0-9]+$",
                                            "type": "string"
                                        },
                                        "minItems": 1,
                                        "type": "array"
                                    }
                                },
                                "required": [
                                    "founded",
                                    "website",
                                    "models"
                                ],
                                "type": "object"
                            }
                        },
                        "required": [
                            "spec"
                        ],
                        "type": "object"
                    }
                },
                "subresources": {
                    "status": {}
                },
                "additionalPrinterColumns": [
                    {
                        "name": "Founded",
                        "type": "string",
                        "jsonPath": ".spec.founded"
                    },
                    {
                        "name": "Age",
                        "type": "date",
                        "jsonPath": ".metadata.creationTimestamp"
                    }
                ]
            }
        ]
    }
}

def test_manufacturer_crd():
    """
    Verifies that the manufacturer CRD resource is as expected.
    """
    crd = registry.get_crd(API_GROUP, "Manufacturer")
    assert crd.kubernetes_resource() == MANUFACTURER_CRD_EXPECTED


CAR_CRD_EXPECTED = {
    "apiVersion": "apiextensions.k8s.io/v1",
    "kind": "CustomResourceDefinition",
    "metadata": {
        "name": "cars.test.kcr.azimuth-cloud.io"
    },
    "spec": {
        "group": "test.kcr.azimuth-cloud.io",
        "scope": "Namespaced",
        "names": {
            "kind": "Car",
            "singular": "car",
            "plural": "cars",
            "shortNames": [],
            "categories": []
        },
        "versions": [
            {
                "name": "v1alpha1",
                "served": True,
                "storage": True,
                "schema": {
                    "openAPIV3Schema": {
                        "description": "Represents a car.",
                        "properties": {
                            "spec": {
                                "description": "Spec for a car.",
                                "properties": {
                                    "manufacturer": {
                                        "description": "The manufacturer of the car.",
                                        "pattern": "^[a-zA-Z0-9-]+$",
                                        "type": "string"
                                    },
                                    "model": {
                                        "description": "The model of the car.",
                                        "pattern": "^[a-zA-Z0-9-]+$",
                                        "type": "string"
                                    },
                                    "engine": {
                                        "anyOf": [
                                            {
                                                "properties": {
                                                    "petrol": {
                                                        "properties": {
                                                            "capacity": {
                                                                "exclusiveMinimum": True,
                                                                "minimum": 0
                                                            }
                                                        },
                                                        "required": [
                                                            "capacity"
                                                        ]
                                                    }
                                                },
                                                "required": [
                                                    "petrol"
                                                ]
                                            },
                                            {
                                                "properties": {
                                                    "diesel": {
                                                        "properties": {
                                                            "capacity": {
                                                                "exclusiveMinimum": True,
                                                                "minimum": 0
                                                            }
                                                        },
                                                        "required": [
                                                            "capacity"
                                                        ]
                                                    }
                                                },
                                                "required": [
                                                    "diesel"
                                                ]
                                            },
                                            {
                                                "properties": {
                                                    "electric": {
                                                        "properties": {
                                                            "maxPowerOutput": {
                                                                "exclusiveMinimum": True,
                                                                "minimum": 0
                                                            }
                                                        },
                                                        "required": [
                                                            "maxPowerOutput"
                                                        ]
                                                    }
                                                },
                                                "required": [
                                                    "electric"
                                                ]
                                            }
                                        ],
                                        "description": "The engine for the car.",
                                        "properties": {
                                            "petrol": {
                                                "description": "Spec for a combustion engine.",
                                                "properties": {
                                                    "capacity": {
                                                        "description": "The capacity of the engine in litres.",
                                                        "exclusiveMinimum": True,
                                                        "minimum": 0,
                                                        "type": "integer"
                                                    }
                                                },
                                                "required": [
                                                    "capacity"
                                                ],
                                                "type": "object"
                                            },
                                            "diesel": {
                                                "description": "Spec for a combustion engine.",
                                                "properties": {
                                                    "capacity": {
                                                        "description": "The capacity of the engine in litres.",
                                                        "exclusiveMinimum": True,
                                                        "minimum": 0,
                                                        "type": "integer"
                                                    }
                                                },
                                                "required": [
                                                    "capacity"
                                                ],
                                                "type": "object"
                                            },
                                            "electric": {
                                                "description": "Spec for an electric engine.",
                                                "properties": {
                                                    "maxPowerOutput": {
                                                        "description": "The maximum power output of the engine in kW.",
                                                        "exclusiveMinimum": True,
                                                        "minimum": 0,
                                                        "type": "integer"
                                                    }
                                                },
                                                "required": [
                                                    "maxPowerOutput"
                                                ],
                                                "type": "object"
                                            }
                                        },
                                        "type": "object"
                                    },
                                    "colour": {
                                        "description": "Enum of possible colours for a car.",
                                        "enum": [
                                            "White",
                                            "Silver",
                                            "Red",
                                            "Black"
                                        ],
                                        "type": "string"
                                    },
                                    "owner": {
                                        "description": "The owner of the car.",
                                        "minLength": 1,
                                        "nullable": True,
                                        "pattern": "^[a-zA-Z0-9 ]*$",
                                        "type": "string"
                                    },
                                    "extras": {
                                        "additionalProperties": {
                                            "x-kubernetes-preserve-unknown-fields": True
                                        },
                                        "description": "Any extras for the car.",
                                        "type": "object",
                                        "x-kubernetes-preserve-unknown-fields": True
                                    }
                                },
                                "required": [
                                    "manufacturer",
                                    "model",
                                    "engine",
                                    "colour"
                                ],
                                "type": "object"
                            },
                            "status": {
                                "description": "Status for a car.",
                                "properties": {
                                    "phase": {
                                        "description": "Enum of possible phases for the car.",
                                        "enum": [
                                            "Unknown",
                                            "OnOrder",
                                            "Building",
                                            "Ready",
                                            "Delayed"
                                        ],
                                        "type": "string"
                                    }
                                },
                                "type": "object",
                                "x-kubernetes-preserve-unknown-fields": True
                            }
                        },
                        "required": [
                            "spec"
                        ],
                        "type": "object"
                    }
                },
                "subresources": {
                    "status": {}
                },
                "additionalPrinterColumns": [
                    {
                        "name": "Manufacturer",
                        "type": "string",
                        "jsonPath": ".spec.manufacturer"
                    },
                    {
                        "name": "Model",
                        "type": "string",
                        "jsonPath": ".spec.model"
                    },
                    {
                        "name": "Phase",
                        "type": "string",
                        "jsonPath": ".status.phase"
                    },
                    {
                        "name": "Age",
                        "type": "date",
                        "jsonPath": ".metadata.creationTimestamp"
                    }
                ]
            }
        ]
    }
}

def test_car_crd():
    """
    Verifies that the car CRD resource is as expected.
    """
    crd = registry.get_crd(API_GROUP, "Car")
    assert crd.kubernetes_resource() == CAR_CRD_EXPECTED



import dataclasses
import importlib
import inspect
import pkgutil
import types
import typing

from .custom_resource import CustomResource, Scope


@dataclasses.dataclass
class CustomResourceDefinitionVersion:
    """
    Storage object for CRD version information.
    """
    # The name of the version
    name: str
    # The model that defines the schema for the version
    model: CustomResource
    # The subresources for the version
    subresources: typing.Dict[str, typing.Any]
    # A list of printer column definitions for the version
    printer_columns: typing.List[typing.Dict[str, typing.Any]]
    # Indicates if this is the storage version
    storage: bool


@dataclasses.dataclass
class CustomResourceDefinition:
    """
    Storage object for CRD information.
    """
    # The API group for the resource
    api_group: str
    # The kind of the resource
    kind: str
    # The singular name of the resource
    singular_name: str
    # The plural name of the resource
    plural_name: str
    # The scope of the resource
    scope: Scope
    # A list of short names for the resource
    short_names: typing.List[str]
    # A list of categories for the resource
    categories: typing.List[str]
    # The versions for the resource, indexed by name
    versions: typing.Dict[str, CustomResourceDefinitionVersion]

    def kubernetes_resource(
        self,
        /,
        include_defaults: bool = False
    ) -> typing.Dict[str, typing.Any]:
        return {
            "apiVersion": "apiextensions.k8s.io/v1",
            "kind": "CustomResourceDefinition",
            "metadata": {
                "name": f"{self.plural_name}.{self.api_group}",
            },
            "spec": {
                "group": self.api_group,
                "scope": self.scope.value,
                "names": {
                    "kind": self.kind,
                    "singular": self.singular_name,
                    "plural": self.plural_name,
                    "shortNames": self.short_names,
                    "categories": self.categories,
                },
                "versions": [
                    {
                        "name": version.name,
                        "served": True,
                        "storage": version.storage,
                        "schema": {
                            "openAPIV3Schema": version.model.model_json_schema(
                                include_defaults = include_defaults
                            ),
                        },
                        "subresources": version.subresources,
                        "additionalPrinterColumns": version.printer_columns,
                    }
                    for version in self.versions.values()
                ]
            }
        }


def iscustomresourcemodel(obj):
    """
    Utility function to check if a given object is a custom resource model.
    """
    return (
        inspect.isclass(obj) and
        issubclass(obj, CustomResource) and
        # Only consider concrete custom resource models
        hasattr(obj, "_meta")
    )


class CustomResourceRegistry:
    """
    Class for indexing and querying available custom resources.
    """
    def __init__(
        self,
        api_group: str,
        categories: typing.Optional[typing.List[str]] = None
    ):
        self._api_group = api_group
        self._categories = categories or []
        # The indexed CRDs indexed by a tuple of (API group, kind)
        self._crds: typing.Dict[typing.Tuple[str, str], CustomResourceDefinition] = {}

    def register_model(self, model: CustomResource):
        """
        Registers the given model as providing a version of a CRD.
        """
        # Combine the root API group with any subgroup for the resource
        api_group = (
            f"{model._meta.api_subgroup}.{self._api_group}"
            if model._meta.api_subgroup
            else self._api_group
        )
        key = (api_group, model._meta.kind)
        # If the CRD exists, we want to extend the short names and the versions
        existing_crd = self._crds.get(key)
        self._crds[key] = CustomResourceDefinition(
            # Last writer wins for resource-level stuff
            api_group,
            model._meta.kind,
            model._meta.singular_name,
            model._meta.plural_name,
            model._meta.scope,
            # Merge but dedupe the short names given by each version
            list(set(getattr(existing_crd, "short_names", []) + model._meta.short_names)),
            self._categories,
            # Add the new version
            dict(
                getattr(existing_crd, "versions", {}),
                **{
                    model._meta.version: CustomResourceDefinitionVersion(
                        model._meta.version,
                        model,
                        model._meta.subresources,
                        model._meta.printer_columns,
                        model._meta.storage_version
                    )
                }
            )
        )
        return model

    def discover_models(self, module: types.ModuleType):
        """
        Discovers and registers all the custom resource models in the given module.
        """
        # Register all the models in this module
        for _, model in inspect.getmembers(module, lambda obj: iscustomresourcemodel(obj)):
            self.register_model(model)
        # Register all the models in the submodules of the given module
        # Only modules with submodules have a __path__, it seems
        if hasattr(module, "__path__"):
            for _, name, _ in pkgutil.iter_modules(module.__path__):
                self.discover_models(importlib.import_module(f".{name}", module.__name__))

    def get_crd(self, api_group, kind) -> typing.Type[CustomResourceDefinition]:
        """
        Returns the CRD definition for the given API group and kind.
        """
        return self._crds[(api_group, kind)]

    def get_model(self, api_group, version, kind) -> typing.Type[CustomResource]:
        """
        Returns the model associated with the given API group, version and kind.
        """
        return self.get_crd(api_group, kind).versions[version].model

    def get_model_instance(self, resource) -> CustomResource:
        """
        Given a Kubernetes resource, return an instance of the model associated with the
        API version and kind.
        """
        api_group, version = resource["apiVersion"].split("/")
        model = self.get_model(api_group, version, resource["kind"])
        return model.model_validate(resource)

    def __iter__(self):
        # Produce the CRDs when iterated
        yield from self._crds.values()

    def __len__(self):
        return len(self._crds)



import copy

import pytest

import pydantic as p

import kube_custom_resource as kcr

import models
from models import v1alpha1 as api


API_GROUP = "test.kcr.azimuth-cloud.io"


registry = kcr.CustomResourceRegistry(API_GROUP)
registry.discover_models(models)


VALID_MANUFACTURER = {
    "apiVersion": f"{API_GROUP}/v1alpha1",
    "kind": "Manufacturer",
    "metadata": {
        "name": "ford",
    },
    "spec": {
        "founded": 1903,
        "website": "https://www.ford.co.uk",
        "models": [
            "escort",
            "sierra",
            "mondeo",
            "fiesta",
            "focus",
        ],
    },
}

def test_validate_valid_manufacturer():
    """
    Tests that a valid manufacturer resource can be loaded.
    """
    manufacturer: api.Manufacturer = registry.get_model_instance(VALID_MANUFACTURER)
    assert manufacturer.api_version == f"{API_GROUP}/v1alpha1"
    assert manufacturer.kind == "Manufacturer"
    assert manufacturer.metadata.name == "ford"
    assert manufacturer.spec.founded == "1903"
    assert manufacturer.spec.website == "https://www.ford.co.uk/"
    assert len(manufacturer.spec.models) == 5


VALID_CAR = {
    "apiVersion": f"{API_GROUP}/v1alpha1",
    "kind": "Car",
    "metadata": {
        "name": "john-doe-escort",
    },
    "spec": {
        "manufacturer": "ford",
        "model": "escort",
        "engine": {
            "petrol": {
                "capacity": 2,
            },
        },
        "colour": "Black",
        "extras": {
            "mats": True,
        },
    },
    "status": {
        "phase": "OnOrder",
    },
}

def test_validate_valid_car():
    """
    Tests that a valid car resource can be loaded.
    """
    car: api.Car = registry.get_model_instance(VALID_CAR)
    assert car.api_version == f"{API_GROUP}/v1alpha1"
    assert car.kind == "Car"
    assert car.metadata.name == "john-doe-escort"
    assert car.spec.manufacturer == "ford"
    assert car.spec.model == "escort"
    assert type(car.spec.engine) == api.PetrolEngine
    assert car.spec.engine.petrol.capacity == 2
    assert car.spec.colour == api.Colour.BLACK
    assert car.spec.owner is None
    assert car.spec.extras == {"mats": True}
    assert car.status.phase == api.Phase.ON_ORDER


def test_validate_invalid_car_raises_validationerror():
    """
    Tests that attempting to load an invalid resource results in a validation error.
    """
    with pytest.raises(p.ValidationError):
        invalid_car = copy.deepcopy(VALID_CAR)
        invalid_car["spec"]["colour"] = "Green"
        _ = registry.get_model_instance(invalid_car)



import enum
import typing as t

import annotated_types

from pydantic_core import CoreSchema, core_schema

from pydantic import (
    BaseModel as PydanticModel,
    TypeAdapter,
    AnyUrl as PydanticAnyUrl,
    AnyHttpUrl as PydanticAnyHttpUrl,
    TypeAdapter,
    StringConstraints,
    AllowInfNan,
    Strict,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler
)
from pydantic.json_schema import JsonSchemaValue


def schema_apply(schema, func, pre = False):
    """
    Return a new schema that is the result of recursively applying the given function
    to the given schema and all subschemas.

    The function itself can be applied to the given schema either before or after
    it is applied to the subschemas.
    """
    schema_new = schema.copy()
    # Check if we need to apply the function before recursing
    if pre:
        schema_new = func(schema_new)
    # For object schemas, we need to apply the function to each of the property schemas
    if schema_new.get("type") == "object":
        if "properties" in schema_new:
            schema_new["properties"] = {
                prop_name: schema_apply(prop_schema, func, pre)
                for prop_name, prop_schema in schema_new["properties"].items()
            }
        # Additional properties can be the boolean "true"
        # If it is a dict, then it is the schema for the additional properties
        additional_properties = schema_new.get("additionalProperties")
        if isinstance(additional_properties, dict):
            schema_new["additionalProperties"] = schema_apply(additional_properties, func, pre)
    # For array schemas, we need to apply the function to the schema for items
    elif schema_new.get("type") == "array":
        if "items" in schema_new:
            schema_new["items"] = schema_apply(schema_new["items"], func, pre)
    # For anyOf, allOf and oneOf, we need to apply the function to each schema
    if "anyOf" in schema_new:
        schema_new["anyOf"] = [schema_apply(s, func, pre) for s in schema_new["anyOf"]]
    if "allOf" in schema_new:
        schema_new["allOf"] = [schema_apply(s, func, pre) for s in schema_new["allOf"]]
    if "oneOf" in schema_new:
        schema_new["oneOf"] = [schema_apply(s, func, pre) for s in schema_new["oneOf"]]
    # not is a single schema
    if "not" in schema_new:
        schema_new["not"] = schema_apply(schema_new["not"], func, pre)
    # Finally, apply the function to the schema itself if required
    return schema_new if pre else func(schema_new)


def resolve_refs(schema):
    """
    Recursively resolve $refs in the given schema.
    """
    schema_new = schema.copy()
    definitions = schema_new.pop("$defs", {})

    def func(schema):
        schema_new = schema.copy()
        # If the schema has an allOf with a single item, collapse it
        if "allOf" in schema_new and len(schema_new["allOf"]) == 1:
            schema_new.update(schema_new.pop("allOf")[0])
        # If the schema has a ref, resolve it
        if "$ref" in schema_new:
            # If the schema is a ref schema, resolve the ref
            ref = schema_new.pop("$ref").removeprefix("#/$defs/")
            schema_new.update(definitions[ref])
        return schema_new

    # Apply the function to each schema _before_ recursing
    # This ensures that refs are recursed into correctly
    return schema_apply(schema_new, func, pre = True)


def remove_fields(schema, *fields):
    """
    Recursively remove the specified fields from all the types in the schema.
    """
    return schema_apply(schema, lambda s: { k: v for k, v in s.items() if k not in fields })


def snake_to_pascal(name):
    """
    Converts a snake case name to pascalCase.
    """
    first, *rest = name.split("_")
    return "".join([first] + [part.capitalize() for part in rest])


class Enum(enum.Enum):
    """
    Enum that does not include a title in the JSON-Schema.
    """
    def __str__(self):
        return str(self.value)

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema.pop("title", None)
        return json_schema


class XKubernetesPreserveUnknownFields:
    """
    Type annotation that adds the x-kubernetes-preserve-unknown-fields property
    to the generated schema.
    """
    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema["x-kubernetes-preserve-unknown-fields"] = True
        return json_schema


Any = t.Annotated[t.Any, XKubernetesPreserveUnknownFields]


KeyT = t.TypeVar("KeyT")
ValueT = t.TypeVar("ValueT")
Dict = t.Annotated[t.Dict[KeyT, ValueT], XKubernetesPreserveUnknownFields]


class XKubernetesIntOrString:
    """
    Type annotation that adds the x-kubernetes-int-or-string property to the generated schema.
    """
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: t.Any,
        handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        # Validate the value as a union of int and str, but convert to a string after
        return core_schema.no_info_after_validator_function(
            str,
            handler.generate_schema(t.Union[str, int])
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        # Replace the anyOf generated by the union with the special Kubernetes property
        json_schema.pop("anyOf", None)
        json_schema["x-kubernetes-int-or-string"] = True
        return json_schema


IntOrString = t.Annotated[str, XKubernetesIntOrString]


def constr(**kwargs):
    return t.Annotated[str, StringConstraints(**kwargs)]


class ValidateStrAs:
    """
    Type annotation that allows a str to be annotated with validation from another type.
    """
    def __init__(self, validation_type):
        self.validation_type = validation_type
        self.type_adapter = TypeAdapter(self.validation_type)

    def __get_pydantic_core_schema__(
        self,
        source_type: t.Any,
        handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        # Validate the value as the given validation type, but convert back to a string after
        json_schema = core_schema.no_info_after_validator_function(
            str,
            handler.generate_schema(self.validation_type)
        )
        return core_schema.json_or_python_schema(
            json_schema = json_schema,
            python_schema = json_schema,
            # When serializing, just use the string
            serialization = core_schema.plain_serializer_function_ser_schema(lambda x: x)
        )


AnyUrl = t.Annotated[str, ValidateStrAs(PydanticAnyUrl)]
AnyHttpUrl = t.Annotated[str, ValidateStrAs(PydanticAnyHttpUrl)]


class _ConvertExclusiveMinMax:
    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        exclusive_min = json_schema.pop("exclusiveMinimum", None)
        if exclusive_min is not None:
            json_schema.update({
                "minimum": exclusive_min,
                "exclusiveMinimum": True,
            })
        exclusive_max = json_schema.pop("exclusiveMaximum", None)
        if exclusive_max is not None:
            json_schema.update({
                "maximum": exclusive_max,
                "exclusiveMaximum": True,
            })
        return json_schema


def conint(
    *,
    strict: t.Optional[bool] = None,
    gt: t.Optional[int] = None,
    ge: t.Optional[int] = None,
    lt: t.Optional[int] = None,
    le: t.Optional[int] = None,
    multiple_of: t.Optional[int] = None
) -> t.Type[int]:
    return t.Annotated[
        int,
        Strict(strict) if strict is not None else None,
        annotated_types.Interval(gt = gt, ge = ge, lt = lt, le = le),
        annotated_types.MultipleOf(multiple_of) if multiple_of is not None else None,
        _ConvertExclusiveMinMax,
    ]


def confloat(
    *,
    strict: t.Optional[bool] = None,
    gt: t.Optional[float] = None,
    ge: t.Optional[float] = None,
    lt: t.Optional[float] = None,
    le: t.Optional[float] = None,
    multiple_of: t.Optional[float] = None,
    allow_inf_nan: t.Optional[bool] = None,
) -> t.Type[float]:
    return t.Annotated[
        float,
        Strict(strict) if strict is not None else None,
        annotated_types.Interval(gt = gt, ge = ge, lt = lt, le = le),
        annotated_types.MultipleOf(multiple_of) if multiple_of is not None else None,
        AllowInfNan(allow_inf_nan) if allow_inf_nan is not None else None,
        _ConvertExclusiveMinMax,
    ]


class Nullable:
    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # We don't want to produce the anyOf schema that Optional would normally produce,
        # as it is not valid OpenAPI
        # Instead, we want to produce the schema for the wrapped type with nullable set
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        any_of = json_schema.pop("anyOf")
        non_null_schema = next(s for s in any_of if s.get("type") != "null")
        json_schema.update(non_null_schema)
        json_schema["nullable"] = True
        return json_schema


OptionalT = t.TypeVar("OptionalT")
Optional = t.Annotated[t.Optional[OptionalT], Nullable]


class StructuralUnion:
    """
    Type annotation for a structural union, i.e. a union with a structural schema.

    See https://kubernetes.io/docs/tasks/extend-kubernetes/custom-resources/custom-resource-definitions/#specifying-a-structural-schema.
    """
    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # We only support structural union schemas for unions (obvs!)
        assert core_schema["type"] == "union"
        # Build the anyOf using the fully-resolved schemas of the individual types
        any_of = []
        properties = {}
        for choice_schema in core_schema.get("choices", []):
            if choice_schema["type"] == "model":
                # For models, we need the fully resolved schema to do the rest of the ops
                # The handler only gives us a ref schema
                adapter = TypeAdapter(choice_schema["cls"])
                choice_json_schema = resolve_refs(adapter.json_schema())
            else:
                choice_json_schema = handler.resolve_ref_schema(handler(choice_schema))
            # In order to qualify as a structural schema, the schema of the union itself
            # must include all the possible properties
            properties.update(choice_json_schema.get("properties", {}))
            # Schemas in anyOf are not permitted to contain particular keys
            choice_json_schema = remove_fields(
                choice_json_schema,
                "type",
                "title",
                "description",
                "default",
                "additionalProperties",
                "nullable",
                "x-kubernetes-preserve-unknown-fields",
            )
            any_of.append(choice_json_schema)
        return {
            "type": "object",
            "properties": properties,
            "anyOf": any_of,
        }


class BaseModel(
    PydanticModel,
    alias_generator = snake_to_pascal,
    populate_by_name = True,
    # Validate any mutations to the model
    frozen = False,
    validate_assignment = True
):
    """
    Base model for use within CRD definitions.
    """
    def model_dump(self, **kwargs):
        # Unless otherwise specified, we want by_alias = True
        kwargs.setdefault("by_alias", True)
        return super().model_dump(**kwargs)

    def model_dump_json(self, **kwargs):
        # Unless otherwise specified, we want by_alias = True
        kwargs.setdefault("by_alias", True)
        return super().model_dump_json(**kwargs)

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = super().__get_pydantic_json_schema__(core_schema, handler)
        #####
        # Post-process the generated schema to make it compatible with a Kubernetes CRD
        #####
        # When extra fields are allowed, stop Kubernetes pruning them
        if cls.model_config.get("extra") == "allow":
            json_schema["x-kubernetes-preserve-unknown-fields"] = True
            json_schema.pop("additionalProperties", None)
        return json_schema

    @classmethod
    def model_json_schema(cls, *args, include_defaults = False, **kwargs):
        json_schema = super().model_json_schema(*args, **kwargs)
        # Resolve all the refs
        json_schema = resolve_refs(json_schema)
        # Remove the titles
        fields_to_remove = ["title"]
        # Unless explicitly included, we remove defaults from the schema as they cause
        # Kubernetes to rewrite objects
        # In most cases, it is better that defaults are applied at model instantiation time
        # as rewriting the Kubernetes objects themselves can have unintended side-effects
        # However in some cases it is more appropriate for the defaults to be "locked in" at
        # creation time
        if not include_defaults:
            fields_to_remove.append("default")
        return remove_fields(json_schema, *fields_to_remove)



import importlib
import importlib.util
import pathlib

import click
import yaml

from .registry import CustomResourceRegistry


@click.command()
@click.option(
    "--category",
    "categories",
    multiple = True,
    help = "Category to place CRDs in. Can be specified multiple times."
)
@click.argument("models_mod")
@click.argument("api_group")
@click.argument("output_dir", type = click.Path())
def main(models_mod, api_group, output_dir, categories = None):
    """
    Generate CRDs from the models in the given module with the given API group.

    The generated CRDs are written into the specified directory.
    """
    # Load the specified module
    models_module = importlib.import_module(models_mod)
    # Create a registry and discover the models in the module
    registry = CustomResourceRegistry(api_group, categories)
    registry.discover_models(models_module)
    # Produce the CRDs in the output directory
    output_dir = pathlib.Path(output_dir)
    output_dir.mkdir(parents = True, exist_ok = True)
    for crd in registry:
        output_path = output_dir / f"{crd.plural_name}.{crd.api_group}.yaml"
        with output_path.open("w") as fh:
            yaml.safe_dump(crd.kubernetes_resource(), fh)



import typing as t

import pydantic as p

from kube_custom_resource import custom_resource as crd, schema as s


class CombustionEngineSpec(s.BaseModel):
    """Spec for a combustion engine."""
    capacity: s.conint(gt = 0) = p.Field(
        ...,
        description = "The capacity of the engine in litres."
    )


class ElectricEngineSpec(s.BaseModel):
    """Spec for an electric engine."""
    max_power_output: s.conint(gt = 0) = p.Field(
        ...,
        description = "The maximum power output of the engine in kW."
    )



class PetrolEngine(s.BaseModel):
    """Union member for a petrol engine."""
    petrol: CombustionEngineSpec


class DieselEngine(s.BaseModel):
    """Union member for a diesel engine."""
    diesel: CombustionEngineSpec


class ElectricEngine(s.BaseModel):
    """Union member for an electric engine."""
    electric: ElectricEngineSpec


Engine = t.Annotated[
    t.Union[PetrolEngine, DieselEngine, ElectricEngine],
    s.StructuralUnion
]


class Colour(str, s.Enum):
    """Enum of possible colours for a car."""
    WHITE  = "White"
    SILVER = "Silver"
    RED    = "Red"
    BLACK  = "Black"


class CarSpec(s.BaseModel):
    """Spec for a car."""
    manufacturer: s.constr(pattern = r"^[a-zA-Z0-9-]+$") = p.Field(
        ...,
        description = "The manufacturer of the car."
    )
    model: s.constr(pattern = r"^[a-zA-Z0-9-]+$") = p.Field(
        ...,
        description = "The model of the car."
    )
    engine: Engine = p.Field(
        ...,
        description = "The engine for the car."
    )
    colour: Colour = p.Field(
        ...,
        description = "The colour of the car."
    )
    owner: s.Optional[
        s.constr(
            pattern = r"^[a-zA-Z0-9 ]*$",
            strip_whitespace = True,
            min_length = 1
        )
    ] = p.Field(
        None,
        description = "The owner of the car."
    )
    extras: s.Dict[str, s.Any] = p.Field(
        default_factory = dict,
        description = "Any extras for the car."
    )


class Phase(str, s.Enum):
    """Enum of possible phases for the car."""
    UNKNOWN  = "Unknown"
    ON_ORDER = "OnOrder"
    BUILDING = "Building"
    READY    = "Ready"
    DELAYED  = "Delayed"


class CarStatus(s.BaseModel, extra = "allow"):
    """Status for a car."""
    phase: Phase = p.Field(
        Phase.UNKNOWN,
        description = "The phase of the car."
    )


class Car(
    crd.CustomResource,
    scope = crd.Scope.NAMESPACED,
    subresources = {"status": {}},
    printer_columns = [
        {
            "name": "Manufacturer",
            "type": "string",
            "jsonPath": ".spec.manufacturer",
        },
        {
            "name": "Model",
            "type": "string",
            "jsonPath": ".spec.model",
        },
        {
            "name": "Phase",
            "type": "string",
            "jsonPath": ".status.phase",
        },
    ],
):
    """Represents a car."""
    spec: CarSpec
    status: CarStatus = p.Field(default_factory = CarStatus)
